@@ -1,7 +1,7 @@
 #
 # This file is part of Config-Model-Tester
 #
-# This software is Copyright (c) 2014 by Dominique Dumont.
+# This software is Copyright (c) 2013-2016 by Dominique Dumont.
 #
 # This is free software, licensed under:
 #
@@ -9,22 +9,18 @@
 #
 package Config::Model::Tester;
 # ABSTRACT: Test framework for Config::Model
-$Config::Model::Tester::VERSION = '2.048';
+$Config::Model::Tester::VERSION = '2.053';
 use warnings;
 use strict;
 use locale;
 use utf8;
+use 5.10.1;
 
 use Test::More;
 use Log::Log4perl 1.11 qw(:easy :levels);
-use File::Path;
-use File::Copy;
+use Path::Tiny;
 use File::Copy::Recursive qw(fcopy rcopy dircopy);
-use File::Find;
 
-use Path::Class 0.29;
-
-use File::Spec ;
 use Test::Warn;
 use Test::Exception;
 use Test::File::Contents ;
@@ -36,6 +32,7 @@ use Test::Memory::Cycle ;
 # creating a build dependency loop.
 eval {
     require Config::Model;
+    require Config::Model::Lister;
     require Config::Model::Value;
     require Config::Model::BackendMgr;
 } ;
@@ -49,20 +46,23 @@ require Exporter;
 $File::Copy::Recursive::DirPerms = 0755;
 
 sub setup_test {
-    my ( $model_test, $t_name, $wr_root, $setup ) = @_;
+    my ( $app_to_test, $t_name, $wr_root, $trace, $setup ) = @_;
 
     # cleanup before tests
-    $wr_root->rmtree();
+    $wr_root->remove_tree();
     $wr_root->mkpath( { mode => 0755 } );
 
-    my $wr_dir    = $wr_root->subdir('test-' . $t_name);
-    my $wr_dir2   = $wr_root->subdir('test-' . $t_name.'-w');
+    my $wr_dir    = $wr_root->child('test-' . $t_name);
+    my $wr_dir2   = $wr_root->child('test-' . $t_name.'-w');
     my $conf_file ;
-    $conf_file = $wr_dir->file($conf_dir,$conf_file_name) if defined $conf_file_name;
+    $conf_file = $wr_dir->child($conf_dir,$conf_file_name)
+        if $conf_dir and $conf_file_name;
+
+    my $ex_dir = path('t')->child('model_tests.d', "$app_to_test-examples");
+    my $ex_data = $ex_dir->child($t_name);
 
-    my $ex_dir = dir('t')->subdir('model_tests.d', "$model_test-examples");
-    my $ex_data = -d $ex_dir->subdir($t_name)->stringify ? $ex_dir->subdir($t_name) : $ex_dir->file($t_name);
     my @file_list;
+
     if ($setup) {
         foreach my $file (keys %$setup) {
             my $map = $setup->{$file} ;
@@ -70,30 +70,36 @@ sub setup_test {
                 = ref ($map) eq 'HASH' ? $map->{$^O} // $map->{default}
                 :                        $map;
             if (not defined $destination_str) {
-                die "$model_test $t_name setup error: cannot find destination for test file $file" ;
+                die "$app_to_test $t_name setup error: cannot find destination for test file $file" ;
             }
-            my $destination = $wr_dir->file($destination_str) ;
+            my $destination = $wr_dir->child($destination_str) ;
             $destination->parent->mkpath( { mode => 0755 }) ;
-            my $data = $ex_data->file($file)->slurp() ;
+            my $data = $ex_data->child($file)->slurp() ;
             $destination->spew( $data );
             @file_list = list_test_files ($wr_dir);
         }
     }
     elsif ( $ex_data->is_dir ) {
         # copy whole dir
-        my $debian_dir = $conf_dir ? $wr_dir->subdir($conf_dir) : $wr_dir ;
+        my $debian_dir = $conf_dir ? $wr_dir->child($conf_dir) : $wr_dir ;
         $debian_dir->mkpath( { mode => 0755 });
+        say "dircopy ". $ex_data->stringify . '->'. $debian_dir->stringify
+            if $trace ;
         dircopy( $ex_data->stringify, $debian_dir->stringify )
           || die "dircopy $ex_data -> $debian_dir failed:$!";
         @file_list = list_test_files ($debian_dir);
     }
-    else {
-
+    elsif ( $ex_data->exists ) {
         # just copy file
+        say "file copy ". $ex_data->stringify . '->'. $conf_file->stringify
+            if $trace ;
         fcopy( $ex_data->stringify, $conf_file->stringify )
           || die "copy $ex_data -> $conf_file failed:$!";
     }
-    ok( 1, "Copied $model_test example $t_name" );
+    else {
+        note ('starting test without original config data, i.e. from scratch');
+    }
+    ok( 1, "Copied $app_to_test example $t_name" );
 
     return ( $wr_dir, $wr_dir2, $conf_file, $ex_data, @file_list );
 }
@@ -105,258 +111,408 @@ sub list_test_files {
     my $debian_dir = shift;
     my @file_list ;
 
-	my $chop = scalar $debian_dir->dir_list();
-	my $scan = sub {
-		my ($child) = @_;
-		return if $child->is_dir ;
-		my @l = $child->components();
-		splice @l,0,$chop;
-		push @file_list, '/'.join('/',@l) ; # build a unix-like path even on windows
-	};
+    my $iter = $debian_dir->iterator({ recurse => 1 });
+    my $debian_str = $debian_dir->stringify;
 
-	$debian_dir->recurse(callback => $scan);
+	while ( my $child = $iter->() ) {
+		next if $child->is_dir ;
+
+		push @file_list, '/' . $child->relative($debian_str)->stringify;
+		#push @file_list, '/'.join('/',@l) ; # build a unix-like path even on windows
+	};
 
     return sort @file_list;
 }
 
+sub write_config_file {
+    my ($conf_dir,$wr_dir,$t) = @_;
+
+    if ($t->{config_file}) {
+        my $file = $conf_dir ? "$conf_dir/" : '';
+        $file .= $t->{config_file} ;
+        $wr_dir->child($file)->parent->mkpath({mode => 0755} ) ;
+    }
+}
+
+sub check_load_warnings {
+    my ($root,$t) = @_ ;
+
+    if ( ($t->{no_warnings} or exists $t->{load_warnings}) and not defined $t->{load_warnings}) {
+        local $Config::Model::Value::nowarning = 1;
+        $root->init;
+        ok( 1,"Read configuration and created instance with init() method without warning check" );
+    }
+    else {
+        warnings_like { $root->init; } $t->{load_warnings},
+            "Read configuration and created instance with init() method with warning check ";
+    }
+}
+
+sub run_update {
+    my ($inst, $dir, $t) = @_;
+    my %args = %{$t->{update}};
+
+    my $ret = delete $args{returns};
+
+    local $Config::Model::Value::nowarning = $args{no_warnings} || $t->{no_warnings} || 0;
+
+    note("updating config with ". join(' ',%args));
+    my $res = $inst->update( from_dir => $dir, %args ) ;
+    if (defined $ret) {
+        is($res,$ret,"updated configuration, got expected return value");
+    }
+    else {
+        ok(1,"updated configuration");
+    }
+}
+
+sub load_instructions {
+    my ($root,$t,$trace) = @_ ;
+
+    print "Loading $t->{load}\n" if $trace ;
+    $root->load( $t->{load} );
+    ok( 1, "load called" );
+}
+
+sub apply_fix {
+    my $inst = shift;
+    local $Config::Model::Value::nowarning = 1;
+    $inst->apply_fixes;
+    ok( 1, "apply_fixes called" );
+}
+
+sub dump_tree_full_mode {
+    my ($app_to_test, $root, $t, $trace) = @_;
+
+    print "dumping tree ...\n" if $trace;
+    my $dump  = '';
+    my $risky = sub {
+        $dump = $root->dump_tree( mode => 'full' );
+    };
+
+    if ( defined $t->{dump_errors} ) {
+        my $nb = 0;
+        my @tf = @{ $t->{dump_errors} };
+        while (@tf) {
+            my $qr = shift @tf;
+            throws_ok { &$risky } $qr, "Failed dump $nb of $app_to_test config tree";
+            my $fix = shift @tf;
+            $root->load($fix);
+            ok( 1, "Fixed error nb " . $nb++ );
+        }
+    }
+
+    if ( ($t->{no_warnings} or (exists $t->{dump_warnings}) and not defined $t->{dump_warnings}) ) {
+        local $Config::Model::Value::nowarning = 1;
+        &$risky;
+        ok( 1, "Ran dump_tree (no warning check)" );
+    }
+    else {
+        warnings_like { &$risky; } $t->{dump_warnings}, "Ran dump_tree";
+    }
+    ok( $dump, "Dumped $app_to_test config tree in full mode" );
+
+    print $dump if $trace;
+    return $dump;
+}
+
+# TODO: factorise with function above and create parameters to
+# handle warnings in both cases with warnings_like
+sub dump_tree_custom_mode {
+    my ($label, $root, $t, $trace) = @_;
+
+    local $Config::Model::Value::nowarning = $t->{no_warnings} || 0;
+
+    my $dump = $root->dump_tree();
+    ok( $dump, "Dumped $label config tree in custom mode" );
+    return $dump;
+}
+
+sub check_data {
+    my ($label, $root, $c, $nw) = @_;
+
+    local $Config::Model::Value::nowarning = $nw || 0;
+    my @checks = ref $c eq 'ARRAY' ? @$c
+        : map { ( $_ => $c->{$_})} sort keys %$c ;
+
+    while (@checks) {
+        my $path       = shift @checks;
+        my $v          = shift @checks;
+        my $check_v    = ref $v eq 'HASH' ? delete $v->{value} : $v;
+        my @check_args = ref $v eq 'HASH' ? %$v : ();
+        my $check_str  = @check_args ? " (@check_args)" : '';
+        my $obj = $root->grab( step => $path, type => ['leaf','check_list'], @check_args );
+        my $got = $obj->fetch(@check_args);
+        if (ref $check_v eq 'Regexp') {
+            like( $got, $check_v, "$label check '$path' value with regexp$check_str" );
+        }
+        else {
+            is( $got, $check_v, "$label check '$path' value$check_str" );
+        }
+    }
+}
+
+sub check_annotation {
+    my ($root, $t) = @_;
+
+    my $annot_check = $t->{verify_annotation};
+    foreach my $path (keys %$annot_check) {
+        my $note = $annot_check->{$path};
+        is( $root->grab($path)->annotation, $note, "check $path annotation" );
+    }
+}
+
+sub has_key {
+    my ($root, $c, $nw) = @_;
+
+    _test_key($root, $c, $nw, 0);
+}
+
+sub has_not_key {
+    my ($root, $c, $nw) = @_;
+
+    _test_key($root, $c, $nw, 1);
+}
+
+sub _test_key {
+    my ($root, $c, $nw, $invert) = @_;
+
+    my @checks = ref $c eq 'ARRAY' ? @$c
+        : map { ( $_ => $c->{$_})} sort keys %$c ;
+
+    while (@checks) {
+        my $path       = shift @checks;
+        my $spec       = shift @checks;
+        my @key_checks = ref $spec eq 'ARRAY' ? @$spec: ($spec);
+
+        my $obj = $root->grab( step => $path, type => 'hash' );
+        my @keys = $obj->fetch_all_indexes;
+        my $res = 0;
+        foreach my $check (@key_checks) {
+            my @match  ;
+            foreach my $k (@keys) {
+                if (ref $check eq 'Regexp') {
+                    push @match, $k if $k =~ $check;
+                }
+                else {
+                    push @match, $k if $k eq $check;
+                }
+            }
+            if ($invert) {
+                is(scalar @match,0, "check $check matched no key" );
+            }
+            else {
+                ok(scalar @match, "check $check matched with keys @match" );
+            }
+        }
+    }
+}
+
+sub write_data_back {
+    my ($app_to_test, $inst, $t) = @_;
+    local $Config::Model::Value::nowarning = $t->{no_warnings} || 0;
+    $inst->write_back( force => 1 );
+    ok( 1, "$app_to_test write back done" );
+}
+
+sub check_file_content {
+    my ($wr_dir, $t) = @_;
+
+    if (my $fc = $t->{file_contents} || $t->{file_content}) {
+        foreach my $f (keys %$fc) {
+            my $t = $fc->{$f} ;
+            my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
+            foreach (@tests) {
+                file_contents_eq_or_diff $wr_dir->child($f)->stringify,  $_,
+                    "check that $f contains $_";
+            }
+        }
+    }
+
+    if (my $fc = $t->{file_contents_like}) {
+        foreach my $f (keys %$fc) {
+            my $t = $fc->{$f} ;
+            my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
+            foreach (@tests) {
+                file_contents_like $wr_dir->child($f)->stringify,  $_,
+                    "check that $f matches regexp $_";
+            }
+        }
+    }
+
+    if (my $fc = $t->{file_contents_unlike}) {
+        foreach my $f (keys %$fc) {
+            my $t = $fc->{$f} ;
+            my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
+            foreach (@tests) {
+                file_contents_unlike $wr_dir->child($f)->stringify,  $_,
+                    "check that $f does not match regexp $_";
+            }
+        }
+    }
+}
+
+sub check_added_or_removed_files {
+    my ( $conf_dir, $wr_dir, $t, @file_list) = @_;
+
+    # copy whole dir
+    my $debian_dir = $conf_dir ? $wr_dir->child($conf_dir) : $wr_dir ;
+    my @new_file_list = list_test_files($debian_dir) ;
+    $t->{file_check_sub}->( \@file_list ) if defined $t->{file_check_sub};
+    eq_or_diff( \@new_file_list, [ sort @file_list ], "check added or removed files" );
+}
+
+sub create_second_instance {
+    my ($app_to_test, $t_name, $wr_dir, $wr_dir2,$t, $config_dir_override) = @_;
+
+    # create another instance to read the conf file that was just written
+    dircopy( $wr_dir->stringify, $wr_dir2->stringify )
+        or die "can't copy from $wr_dir to $wr_dir2: $!";
+
+    my $i2_test = $model->instance(
+        root_class_name => $model_to_test,
+        root_dir        => $wr_dir2->stringify,
+        config_file     => $t->{config_file} ,
+        instance_name   => "$app_to_test-$t_name-w",
+        application     => $app_to_test,
+        check           => $t->{load_check2} || 'yes',
+        config_dir      => $config_dir_override,
+    );
+
+    ok( $i2_test, "Created instance $app_to_test-test-$t_name-w" );
+
+    local $Config::Model::Value::nowarning = $t->{no_warnings} || 0;
+    my $i2_root = $i2_test->config_root;
+    $i2_root->init;
+
+    return $i2_root;
+}
+
 sub run_model_test {
-    my ($model_test, $model_test_conf, $do, $model, $trace, $wr_root) = @_ ;
+    my ($app_to_test, $app_to_test_conf, $do, $model, $trace, $wr_root) = @_ ;
 
     $skip = 0;
     undef $conf_file_name ;
     undef $conf_dir ;
     undef $home_for_test ;
 
-    note("Beginning $model_test test ($model_test_conf)");
+    note("Beginning $app_to_test test ($app_to_test_conf)");
 
-    unless ( my $return = do $model_test_conf ) {
-        warn "couldn't parse $model_test_conf: $@" if $@;
-        warn "couldn't do $model_test_conf: $!" unless defined $return;
-        warn "couldn't run $model_test_conf" unless $return;
+    unless ( my $return = do $app_to_test_conf ) {
+        warn "couldn't parse $app_to_test_conf: $@" if $@;
+        warn "couldn't do $app_to_test_conf: $!" unless defined $return;
+        warn "couldn't run $app_to_test_conf" unless $return;
     }
 
     if ($skip) {
-        note("Skipped $model_test test ($model_test_conf)");
+        note("Skipped $app_to_test test ($app_to_test_conf)");
         return;
     }
 
+    my ($trash, $appli_info, $applications) = Config::Model::Lister::available_models(1);
+
     # even undef, this resets the global variable there
     Config::Model::BackendMgr::_set_test_home($home_for_test) ;
 
-    my $note ="$model_test uses $model_to_test model";
+    if (not defined $model_to_test) {
+        $model_to_test = $applications->{$app_to_test};
+        if (not defined $model_to_test) {
+            my @k = sort values %$applications;
+            my @files = map { $_->{_file} // 'unknown' } values %$appli_info ;
+            die "Cannot find model name for $app_to_test in files >@files<. Know dev models are >@k<. ".
+                "Check your test name (the file ending with -test-conf.pl) or set the \$model_to_test global variable\n";
+        }
+    }
+
+    my $config_dir_override = $appli_info->{$app_to_test}{config_dir}; # may be undef
+
+    my $note ="$app_to_test uses $model_to_test model";
     $note .= " on file $conf_file_name" if defined $conf_file_name;
     note($note);
 
     my $idx = 0;
     foreach my $t (@tests) {
         my $t_name = $t->{name} || "t$idx";
-        if ( defined $do and $do ne $t_name ) {
+        if ( defined $do and $t_name !~ /$do/) {
             $idx++;
             next;
         }
-        note("Beginning subtest $model_test $t_name");
+        note("Beginning subtest $app_to_test $t_name");
 
         my ($wr_dir, $wr_dir2, $conf_file, $ex_data, @file_list)
-            = setup_test ($model_test, $t_name, $wr_root,$t->{setup});
+            = setup_test ($app_to_test, $t_name, $wr_root,$trace, $t->{setup});
 
-        if ($t->{config_file}) {
-            $wr_dir->file($t->{config_file})->parent->mkpath({mode => 0755} ) ;
-        }
+        write_config_file($conf_dir,$wr_dir,$t);
 
         my $inst = $model->instance(
             root_class_name => $model_to_test,
             root_dir        => $wr_dir->stringify,
-            instance_name   => "$model_test-" . $t_name,
+            instance_name   => "$app_to_test-" . $t_name,
+            application     => $app_to_test,
             config_file     => $t->{config_file} ,
             check           => $t->{load_check} || 'yes',
+            config_dir      => $config_dir_override,
         );
 
         my $root = $inst->config_root;
 
-        if ( exists $t->{load_warnings}
-            and not defined $t->{load_warnings} )
-        {
-            local $Config::Model::Value::nowarning = 1;
-            $root->init;
-            ok( 1,"Read configuration and created instance with init() method without warning check" );
-        }
-        else {
-            warnings_like { $root->init; } $t->{load_warnings},
-                "Read configuration and created instance with init() method with warning check ";
-        }
+        check_load_warnings ($root,$t);
 
-        if ( $t->{load} ) {
-            print "Loading $t->{load}\n" if $trace ;
-            $root->load( $t->{load} );
-            ok( 1, "load called" );
-        }
+        run_update($inst,$wr_dir,$t) if $t->{update};
 
-        if ( $t->{apply_fix} ) {
-            local $Config::Model::Value::nowarning = 1;
-            $inst->apply_fixes;
-            ok( 1, "apply_fixes called" );
-        }
+        load_instructions ($root,$t,$trace) if $t->{load} ;
 
-        print "dumping tree ...\n" if $trace;
-        my $dump  = '';
-        my $risky = sub {
-            $dump = $root->dump_tree( mode => 'full' );
-        };
+        apply_fix($inst) if  $t->{apply_fix};
 
-        if ( defined $t->{dump_errors} ) {
-            my $nb = 0;
-            my @tf = @{ $t->{dump_errors} };
-            while (@tf) {
-                my $qr = shift @tf;
-                throws_ok { &$risky } $qr,
-                  "Failed dump $nb of $model_test config tree";
-                my $fix = shift @tf;
-                $root->load($fix);
-                ok( 1, "Fixed error nb " . $nb++ );
-            }
-        }
+        dump_tree_full_mode ($app_to_test, $root, $t, $trace) ;
 
-        if ( exists $t->{dump_warnings}
-            and not defined $t->{dump_warnings} )
-        {
-            local $Config::Model::Value::nowarning = 1;
-            &$risky;
-            ok( 1, "Ran dump_tree (no warning check)" );
-        }
-        else {
-            warnings_like { &$risky; } $t->{dump_warnings}, "Ran dump_tree";
-        }
-        ok( $dump, "Dumped $model_test config tree in full mode" );
+        my $dump = dump_tree_custom_mode ($app_to_test, $root, $t, $trace) ;
 
-        print $dump if $trace;
+        check_data("first", $root, $t->{check}, $t->{no_warnings}) if $t->{check};
 
-        local $Config::Model::Value::nowarning = $t->{no_warnings} || 0;
+        has_key     ( $root, $t->{has_key}, $t->{no_warnings}) if $t->{has_key} ;
+        has_not_key ( $root, $t->{has_not_key}, $t->{no_warnings}) if $t->{has_not_key} ;
 
-        $dump = $root->dump_tree();
-        ok( $dump, "Dumped $model_test config tree in custom mode" );
+        check_annotation($root,$t) if $t->{verify_annotation};
 
-        my $c = $t->{check} || {};
-        my @checks = ref $c eq 'ARRAY' ? @$c
-            : map { ( $_ => $c->{$_})} sort keys %$c ;
-        while (@checks) {
-            my $path       = shift @checks;
-            my $v          = shift @checks;
-            my $check_v    = ref $v ? delete $v->{value} : $v;
-            my @check_args = ref $v ? %$v : ();
-            my $check_str  = @check_args ? " (@check_args)" : '';
-            is( $root->grab( step => $path, @check_args )->fetch(@check_args),
-                $check_v, "check '$path' value$check_str" );
-        }
+        write_data_back ($app_to_test, $inst, $t) ;
 
-        if (my $annot_check = $t->{verify_annotation}) {
-            foreach my $path (keys %$annot_check) {
-                my $note = $annot_check->{$path};
-                is( $root->grab($path)->annotation,
-                    $note, "check $path annotation" );
-            }
-        }
+        check_file_content($wr_dir,$t) ;
 
-        $inst->write_back( force => 1 );
-        ok( 1, "$model_test write back done" );
+        check_added_or_removed_files ($conf_dir, $wr_dir, $t, @file_list) if $ex_data->is_dir;
 
-        if (my $fc = $t->{file_contents} || $t->{file_content}) {
-            foreach my $f (keys %$fc) {
-                my $t = $fc->{$f} ;
-                my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
-                foreach (@tests) {
-                    file_contents_eq_or_diff $wr_dir->file($f)->stringify,  $_,
-                        "check that $f contains $_";
-                }
-            }
-        }
+        my $i2_root = create_second_instance ($app_to_test, $t_name, $wr_dir, $wr_dir2,$t, $config_dir_override);
 
-        if (my $fc = $t->{file_contents_like}) {
-            foreach my $f (keys %$fc) {
-                my $t = $fc->{$f} ;
-                my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
-                foreach (@tests) {
-                    file_contents_like $wr_dir->file($f)->stringify,  $_,
-                        "check that $f matches regexp $_";
-                }
-            }
-        }
+        my $p2_dump = dump_tree_custom_mode("second $app_to_test", $i2_root, $t) ;
 
-        if (my $fc = $t->{file_contents_unlike}) {
-            foreach my $f (keys %$fc) {
-                my $t = $fc->{$f} ;
-                my @tests = ref $t eq 'ARRAY' ? @$t : ($t) ;
-                foreach (@tests) {
-                    file_contents_unlike $wr_dir->file($f)->stringify,  $_,
-                        "check that $f does not match regexp $_";
-                }
-            }
-        }
-
-        my @new_file_list;
-        if ( $ex_data->is_dir ) {
-
-            # copy whole dir
-            my $debian_dir = $conf_dir ? $wr_dir->subdir($conf_dir) : $wr_dir ;
-            my @new_file_list = list_test_files($debian_dir) ;
-            $t->{file_check_sub}->( \@file_list )
-              if defined $t->{file_check_sub};
-            eq_or_diff( \@new_file_list, [ sort @file_list ],
-                "check added or removed files" );
-        }
-
-        # create another instance to read the conf file that was just written
-        dircopy( $wr_dir->stringify, $wr_dir2->stringify )
-          or die "can't copy from $wr_dir to $wr_dir2: $!";
-
-        my $i2_test = $model->instance(
-            root_class_name => $model_to_test,
-            root_dir        => $wr_dir2->stringify,
-            config_file     => $t->{config_file} ,
-            instance_name   => "$model_test-$t_name-w",
-            check           => $t->{load_check2} || 'yes',
-        );
-
-        ok( $i2_test, "Created instance $model_test-test-$t_name-w" );
-
-        my $i2_root = $i2_test->config_root;
-        $i2_root->init;
-
-        my $p2_dump = $i2_root->dump_tree();
-        ok( $dump, "Dumped $model_test 2nd config tree in custom mode" );
-
-        eq_or_diff( $p2_dump, $dump,
-            "compare original $model_test custom data with 2nd instance custom data"
+        unified_diff;
+        eq_or_diff(
+            [ split /\n/,$p2_dump ],
+            [ split /\n/,$dump ],
+            "compare original $app_to_test custom data with 2nd instance custom data",
         );
 
         ok( -s "$wr_dir2/$conf_dir/$conf_file_name" ,
-            "check that original $model_test file was not clobbered" )
+            "check that original $app_to_test file was not clobbered" )
                 if defined $conf_file_name ;
 
-        my $wr_check = $t->{wr_check} || {};
-        foreach my $path ( sort keys %$wr_check ) {
-            my $v          = $wr_check->{$path};
-            my $check_v    = ref $v ? delete $v->{value} : $v;
-            my @check_args = ref $v ? %$v : ();
-            is( $i2_root->grab( step => $path, @check_args )->fetch(@check_args),
-                $check_v, "wr_check $path value (@check_args)" );
-        }
+        check_data("second", $i2_root, $t->{wr_check}, $t->{no_warnings}) if $t->{wr_check} ;
 
-        note("End of subtest $model_test $t_name");
+        note("End of subtest $app_to_test $t_name");
 
         $idx++;
     }
-    note("End of $model_test test");
+    note("End of $app_to_test test");
 
 }
 
 sub run_tests {
-    my ( $arg, $test_only_model, $do ) = @_;
+    my ( $arg, $test_only_app, $do ) = @_;
 
-    my ( $log, $show ) = (0) x 2;
+    my $log = 0;
 
     my $trace = ($arg =~ /t/) ? 1 : 0;
     $log  = 1 if $arg =~ /l/;
-    $show = 1 if $arg =~ /s/;
 
     my $log4perl_user_conf_file = ($ENV{HOME} || '') . '/.log4config-model';
 
@@ -378,14 +534,14 @@ sub run_tests {
     ok( 1, "compiled" );
 
     # pseudo root where config files are written by config-model
-    my $wr_root = dir('wr_root');
+    my $wr_root = path('wr_root');
 
     my @group_of_tests = grep { /-test-conf.pl$/ } glob("t/model_tests.d/*");
 
-    foreach my $model_test_conf (@group_of_tests) {
-        my ($model_test) = ( $model_test_conf =~ m!\.d/([\w\-]+)-test-conf! );
-        next if ( $test_only_model and $test_only_model ne $model_test ) ;
-        run_model_test($model_test, $model_test_conf, $do, $model, $trace, $wr_root) ;
+    foreach my $app_to_test_conf (@group_of_tests) {
+        my ($app_to_test) = ( $app_to_test_conf =~ m!\.d/([\w\-]+)-test-conf! );
+        next if ( $test_only_app and $test_only_app ne $app_to_test ) ;
+        run_model_test($app_to_test, $app_to_test_conf, $do, $model, $trace, $wr_root) ;
     }
 
     memory_cycle_ok($model,"test memory cycle") ;
@@ -407,7 +563,7 @@ Config::Model::Tester - Test framework for Config::Model
 
 =head1 VERSION
 
-version 2.048
+version 2.053
 
 =head1 SYNOPSIS
 
@@ -418,11 +574,11 @@ version 2.048
  use Config::Model::Tester ;
  use ExtUtils::testlib;
 
- my $arg = shift || '';
- my $test_only_model = shift || '';
- my $do = shift ;
+ my $arg = shift || ''; # typically e t l
+ my $test_only_app = shift || ''; # only run one set of test
+ my $do = shift ; # select subtests to run with a regexp
 
- run_tests($arg, $test_only_model, $do) ;
+ run_tests($arg, $test_only_app, $do) ;
 
 =head1 DESCRIPTION
 
@@ -430,25 +586,31 @@ This class provides a way to test configuration models with tests files.
 This class was designed to tests several models and several tests
 cases per model.
 
-A specific layout for test files must be followed
+A specific layout for test files must be followed.
 
 =head2 Simple test file layout
 
  t
  |-- model_test.t
- \-- model_tests.d
-     |-- lcdd-test-conf.pl   # test specification
+ \-- model_tests.d           # do not change directory name
+     |-- lcdd-test-conf.pl   # test specification for lcdd model
      \-- lcdd-examples
-         |-- t0              # test case t0
-         \-- LCDD-0.5.5      # test case for older LCDproc
+         |-- t0              # subtest t0
+         \-- LCDD-0.5.5      # subtest for older LCDproc
 
 In the example above, we have 1 model to test: C<lcdd> and 2 tests
-cases.
+cases. The model name matches the file specified in
+C<lib/Config/Model/*.d> directory. In this case, the model name matches 
+C<lib/Config/Model/system.d/lcdproc>
 
-Test specification is written in C<lcdd-test-conf.pl> file. Test
-cases are plain files in C<lcdd-examples>. C<lcdd-test-conf.pl> will
-contain instructions so that each file will be used as a
-C</etc/LCDd.conf> file during each test case.
+Test specification is written in C<lcdd-test-conf.pl> file (i.e. this
+modules looks for files named  like C<< <model-name>-test-conf.pl> >>).
+
+Subtests are specified in files in directory C<lcdd-examples> (
+i.e. this modules looks for subtests in directory 
+C<< <model-name>-examples.pl> >>. C<lcdd-test-conf.pl> contains
+instructions so that each file will be used as a C</etc/LCDd.conf>
+file during each test case.
 
 C<lcdd-test-conf.pl> can contain specifications for more test
 case. Each test case will require a new file in C<lcdd-examples>
@@ -464,7 +626,7 @@ provided in sub-directories:
  t/model_tests.d
  \-- dpkg-test-conf.pl         # test specification
  \-- dpkg-examples
-     \-- libversion            # example subdir
+     \-- libversion            # example subdir, used as subtest name
          \-- debian            # directory for one test case
              |-- changelog
              |-- compat
@@ -477,7 +639,7 @@ provided in sub-directories:
 
 In the example above, the test specification is written in
 C<dpkg-test-conf.pl>. Dpkg layout requires several files per test case.
-C<dpkg-test-conf.pl> will contain instruction so that each directory
+C<dpkg-test-conf.pl> will contain instructions so that each directory
 under C<dpkg-examples> will be used.
 
 See L</Examples> for a link to the (many) Dpkg model tests
@@ -515,12 +677,12 @@ See the actual L<Ssh and Sshd model tests|https://github.com/dod38fr/config-mode
 =head2 Basic test specification
 
 Each model test is specified in C<< <model>-test-conf.pl >>. This file
-contains a set of global variable. (yes, global variables are often bad ideas
+contains a set of global variables. (yes, global variables are often bad ideas
 in programs, but they are handy for tests):
 
  # config file name (used to copy test case into test wr_root directory)
  $conf_file_name = "fstab" ;
- # config dir where to copy the file
+ # config dir where to copy the file (optional)
  #$conf_dir = "etc" ;
  # home directory for this test
  $home_for_test = '/home/joe' ;
@@ -530,7 +692,7 @@ Here, C<t0> file will be copied in C<wr_root/test-t0/etc/fstab>.
  # config model name to test
  $model_to_test = "Fstab" ;
 
- # list of tests
+ # list of tests. This modules looks for @tests global variable
  @tests = (
     {
      # test name
@@ -540,11 +702,16 @@ Here, C<t0> file will be copied in C<wr_root/test-t0/etc/fstab>.
     {
      name => 't1',
      # add optional specification here for t1 test
-     },
+    },
  );
 
  1; # to keep Perl happy
 
+You can suppress warnings by specifying C<< no_warnings => 1 >>. On
+the other hand, you may also want to check for warnings specified to
+your model. In this case, you should avoid specifying C<no_warnings>
+here and specify warning tests or warning filters as mentioned below.
+
 See actual L<fstab test|https://github.com/dod38fr/config-model/blob/master/t/model_tests.d/fstab-test-conf.pl>.
 
 =head2 Internal tests
@@ -577,22 +744,16 @@ See actual L<multistrap test|https://github.com/dod38fr/config-model/blob/master
 =head2 Test scenario
 
 Each subtest follow a sequence explained below. Each step of this
-sequence may be altered by adding specification in the test case:
+sequence may be altered by adding specification in
+C<< <model-to-test>-test-conf.pl >>:
 
 =over
 
 =item *
 
 Setup test in C<< wr_root/<subtest name>/ >>. If your configuration file layout depend
-on the target system, you will have to specify the path using C<setup> parameter:
-
- setup => {
-    'file_name_in_examples_dir' => {
-        'darwin' => '/etc/foo', # macosx
-        'default' => '/etc/bar' # others
-    },
-    'another_file_in_examples_dir' => $computed_path
- }
+on the target system, you will have to specify the path using C<setup> parameter.
+See L</"Test file layout depending on system">.
 
 =item *
 
@@ -606,7 +767,17 @@ E.g.
 
     load_warnings => [ qr/Missing/, (qr/deprecated/) x 3 , ],
 
-Use an empty array_ref to masks load warnings.
+Use an empty array_ref to mask load warnings.
+
+=item *
+
+Optionally run L<update|App::Cme::Command::update> command:
+
+    update => { in => 'some-test-data.txt', returns => 'foo' , no_warnings => [ 0 | 1 ] }
+
+C<returns> is the expected return value (optional). All other
+arguments are passed to C<update> method. Note that C<< quiet => 1 >>
+may be useful for less verbose test.
 
 =item *
 
@@ -615,6 +786,8 @@ suppress any error or warning mentioned above. E.g:
 
     load => 'binary:seaview Synopsis="multiplatform interface for sequence alignment"',
 
+See L<Config::Model::Loader> for the syntax of the string accepted by C<load> parameter.
+
 =item *
 
 Optionally, call L<apply_fixes|Config::Model::Instance/apply_fixes>:
@@ -647,23 +820,52 @@ You can tolerate any dump warning this way:
 Run specific content check to verify that configuration data was retrieved
 correctly:
 
-    check => [
+    check => {
         'fs:/proc fs_spec',           "proc" ,
         'fs:/proc fs_file',           "/proc" ,
         'fs:/home fs_file',          "/home",
-    ],
+    },
+
+The keys of the hash points to the value to be checked using the
+syntax described in L<Config::Model::AnyThing:/"grab(...)">.
 
 You can run check using different check modes (See L<Config::Model::Value/"fetch( ... )">)
 by passing a hash ref instead of a scalar :
 
-    check  => [
-        'sections:debian packages:0' , { qw/mode layered value dpkg-dev/},
-        'sections:base packages:0',    { qw/mode layered value gcc-4.2-base/},
-    ],
+    check  => {
+        'sections:debian packages:0' , { mode => 'layered', value => 'dpkg-dev' },
+        'sections:base packages:0',    { mode => 'layered', value => "gcc-4.2-base' },
+    },
 
 The whole hash content (except "value") is passed to  L<grab|Config::Model::AnyThing/"grab(...)">
 and L<fetch|Config::Model::Value/"fetch( ... )">
 
+A regexp can also be used to check value:
+
+   check => {
+      "License text" => qr/gnu/i,
+      "License text" => { mode => 'custom', value => qr/gnu/i },
+   }
+
+=item *
+
+Verify if a hash contains one or more keys (or keys matching a regexp):
+
+ has_key => [
+    'sections' => 'debian', # sections must point to a hash element
+    'control' => [qw/source binary/],
+    'copyright Files' => qr/.c$/,
+    'copyright Files' => [qr/\.h$/], qr/\.c$/],
+ ],
+
+=item *
+
+Verify that a hash has B<not> a key (or a key matching a regexp):
+
+ has_not_key => [
+    'copyright Files' => qr/.virus$/ # silly, isn't ?
+ ],
+
 =item *
 
 Verify annotation extracted from the configuration file comments:
@@ -679,7 +881,7 @@ Write back the config data in C<< wr_root/<subtest name>/ >>.
 Note that write back is forced, so the tested configuration files are
 written back even if the configuration values were not changed during the test.
 
-You can skip warning when writing back with:
+You can skip warning when writing back with the global :
 
     no_warnings => 1,
 
@@ -739,9 +941,9 @@ Run specific content check on the B<written> config file to verify that
 configuration data was written and retrieved correctly:
 
     wr_check => {
-        'fs:/proc fs_spec',           "proc" ,
-        'fs:/proc fs_file',           "/proc" ,
-        'fs:/home fs_file',          "/home",
+        'fs:/proc fs_spec' =>          "proc" ,
+        'fs:/proc fs_file' =>          "/proc",
+        'fs:/home fs_file' =>          "/home",
     },
 
 Like the C<check> item explained above, you can run C<wr_check> using
@@ -749,11 +951,12 @@ different check modes.
 
 =back
 
-=head2 running the test
+=head2 Running the test
 
-Run all tests:
+Run all tests with one of these commands:
 
- prove -l t/model_test.t
+ prove -l t/model_test.t :: [ t|l|e [ <model_name> [ <regexp> ]]]
+ perl -Ilib t/model_test.t  [ t|l|e [ <model_name> [ <regexp> ]]]
 
 By default, all tests are run on all models.
 
@@ -778,10 +981,13 @@ The model name to tests. E.g.:
 
 =item *
 
-The required subtest E.g.:
+A regexp to filter subtest E.g.:
 
-  # run only fstab tests t0
-  prove -lv t/model_test.t :: x fstab t0
+  # run only fstab tests foobar subtest
+  prove -lv t/model_test.t :: x fstab foobar
+
+  # run only fstab tests foo subtest
+  prove -lv t/model_test.t :: x fstab '^foo$'
 
 =back
 
@@ -839,7 +1045,7 @@ Dominique Dumont
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2014 by Dominique Dumont.
+This software is Copyright (c) 2013-2016 by Dominique Dumont.
 
 This is free software, licensed under:
 
@@ -858,14 +1064,6 @@ in addition to those websites please use your favorite search engine to discover
 
 =item *
 
-MetaCPAN
-
-A modern, open-source CPAN search engine, useful to view POD in HTML format.
-
-L<http://metacpan.org/release/Config-Model-Tester>
-
-=item *
-
 Search CPAN
 
 The default CPAN search engine, useful to view POD in HTML format.
@@ -874,14 +1072,6 @@ L<http://search.cpan.org/dist/Config-Model-Tester>
 
 =item *
 
-RT: CPAN's Bug Tracker
-
-The RT ( Request Tracker ) website is the default bug/issue tracking system for CPAN.
-
-L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Config-Model-Tester>
-
-=item *
-
 AnnoCPAN
 
 The AnnoCPAN is a website that allows community annotations of Perl module documentation.
@@ -898,14 +1088,6 @@ L<http://cpanratings.perl.org/d/Config-Model-Tester>
 
 =item *
 
-CPAN Forum
-
-The CPAN Forum is a web forum for discussing Perl modules.
-
-L<http://cpanforum.com/dist/Config-Model-Tester>
-
-=item *
-
 CPANTS
 
 The CPANTS is a website that analyzes the Kwalitee ( code metrics ) of a distribution.
@@ -940,8 +1122,8 @@ L<http://deps.cpantesters.org/?module=Config::Model::Tester>
 
 =head2 Bugs / Feature Requests
 
-Please report any bugs or feature requests by email to C<bug-config-model-tester at rt.cpan.org>, or through
-the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Config-Model-Tester>. You will be automatically notified of any
+Please report any bugs or feature requests by email to C<ddumont at cpan.org>, or through
+the web interface at L<https://github.com/dod38fr/config-model-tester/issues>. You will be automatically notified of any
 progress on the request by the system.
 
 =head2 Source Code
